@@ -11,38 +11,31 @@ from collections.abc import (
     Iterator,
     Mapping,
     MutableMapping,
-    Set,
 )
 from contextlib import ContextDecorator, contextmanager, suppress
 from dataclasses import dataclass, field
-from enum import Enum, auto
-from functools import (
-    partialmethod,
-    singledispatchmethod,
-    update_wrapper,
-    wraps,
-)
+from functools import partialmethod, singledispatchmethod, update_wrapper
 from inspect import Signature, isclass
-from types import UnionType
+from types import MethodType, UnionType
 from typing import (
     Any,
     ClassVar,
     ContextManager,
+    Literal,
     NamedTuple,
     NoReturn,
     Protocol,
     TypeVar,
     cast,
+    get_args,
     runtime_checkable,
 )
 
 from injection.common.event import Event, EventChannel, EventListener
+from injection.common.invertible import Invertible, SimpleInvertible
 from injection.common.lazy import Lazy, LazyMapping
-from injection.common.tools.threading import (
-    frozen_collection,
-    synchronized,
-    thread_lock,
-)
+from injection.common.queue import LimitedQueue
+from injection.common.tools.threading import synchronized
 from injection.common.tools.type import find_types, format_type, get_origins
 from injection.exceptions import (
     InjectionError,
@@ -52,12 +45,21 @@ from injection.exceptions import (
     NoInjectable,
 )
 
-__all__ = ("Injectable", "Module", "ModulePriority")
+__all__ = ("Injectable", "Module")
 
 _logger = logging.getLogger(__name__)
 
 _T = TypeVar("_T")
-Types = Iterable[type] | UnionType
+_Types = Iterable[type] | UnionType
+
+
+"""
+Literal types
+"""
+
+
+_Mode = Literal["fallback", "normal", "override"]
+_Priority = Literal["low", "high"]
 
 
 """
@@ -73,7 +75,7 @@ class ContainerEvent(Event, ABC):
 @dataclass(frozen=True, slots=True)
 class ContainerDependenciesUpdated(ContainerEvent):
     classes: Collection[type]
-    override: bool
+    mode: _Mode
 
     def __str__(self) -> str:
         length = len(self.classes)
@@ -127,11 +129,11 @@ class ModuleRemoved(ModuleEvent):
 @dataclass(frozen=True, slots=True)
 class ModulePriorityUpdated(ModuleEvent):
     module_updated: Module
-    priority: ModulePriority
+    priority: _Priority
 
     def __str__(self) -> str:
         return (
-            f"In `{self.on_module}`, the priority `{self.priority.name}` "
+            f"In `{self.on_module}`, the priority `{self.priority}` "
             f"has been applied to `{self.module_updated}`."
         )
 
@@ -192,7 +194,7 @@ class SingletonInjectable(BaseInjectable[_T]):
         with suppress(KeyError):
             return self.cache[self.__INSTANCE_KEY]
 
-        with thread_lock:
+        with synchronized():
             instance = self.factory()
             self.cache[self.__INSTANCE_KEY] = instance
 
@@ -203,9 +205,6 @@ class SingletonInjectable(BaseInjectable[_T]):
 class ShouldBeInjectable(Injectable[_T]):
     cls: type[_T]
 
-    def __bool__(self) -> bool:
-        return False
-
     def get_instance(self) -> NoReturn:
         raise InjectionError(f"`{format_type(self.cls)}` should be an injectable.")
 
@@ -242,57 +241,60 @@ Container
 """
 
 
+class Record(NamedTuple):
+    injectable: Injectable
+    mode: _Mode
+
+
 @dataclass(repr=False, frozen=True, slots=True)
 class Container(Broker):
-    __data: dict[type, Injectable] = field(default_factory=dict, init=False)
+    __data: dict[type, Record] = field(default_factory=dict, init=False)
     __channel: EventChannel = field(default_factory=EventChannel, init=False)
 
     def __getitem__(self, cls: type[_T] | UnionType, /) -> Injectable[_T]:
-        for origin in get_origins(cls):
-            with suppress(KeyError):
-                return self.__data[origin]
+        for cls in get_origins(cls):
+            try:
+                injectable, _ = self.__data[cls]
+            except KeyError:
+                continue
+
+            return injectable
 
         raise NoInjectable(cls)
 
     def __contains__(self, cls: type | UnionType, /) -> bool:
-        return any(origin in self.__data for origin in get_origins(cls))
+        return any(cls in self.__data for cls in get_origins(cls))
 
     @property
     def is_locked(self) -> bool:
         return any(injectable.is_locked for injectable in self.__injectables)
 
-    @property
-    def __classes(self) -> frozenset[type]:
-        return frozenset(self.__data.keys())
-
     @property
     def __injectables(self) -> frozenset[Injectable]:
-        return frozenset(self.__data.values())
+        return frozenset(injectable for injectable, _ in self.__data.values())
 
-    def update(self, classes: Iterable[type], injectable: Injectable, override: bool):
-        classes = frozenset(get_origins(*classes))
+    @synchronized()
+    def update(self, classes: Iterable[type], injectable: Injectable, mode: _Mode):
+        records = {
+            cls: Record(injectable, mode)
+            for cls in self.__filter_classes(classes, mode)
+        }
 
-        with thread_lock:
-            if not injectable:
-                classes -= self.__classes
-                override = True
+        if records:
+            event = ContainerDependenciesUpdated(self, records.keys(), mode)
 
-            if classes:
-                event = ContainerDependenciesUpdated(self, classes, override)
-
-                with self.notify(event):
-                    if not override:
-                        self.__check_if_exists(classes)
-
-                    self.__data.update((cls, injectable) for cls in classes)
+            with self.notify(event):
+                self.__data.update(records)
 
         return self
 
-    @synchronized
+    @synchronized()
     def unlock(self):
         for injectable in self.__injectables:
             injectable.unlock()
 
+        return self
+
     def add_listener(self, listener: EventListener):
         self.__channel.add_listener(listener)
         return self
@@ -300,14 +302,27 @@ class Container(Broker):
     def notify(self, event: Event) -> ContextManager | ContextDecorator:
         return self.__channel.dispatch(event)
 
-    def __check_if_exists(self, classes: Set[type]):
-        intersection = classes & self.__classes
+    def __filter_classes(self, classes: Iterable[type], mode: _Mode) -> Iterator[type]:
+        modes = get_args(_Mode)
+        rank = modes.index(mode)
 
-        for cls in intersection:
-            if self.__data[cls]:
-                raise RuntimeError(
-                    f"An injectable already exists for the class `{format_type(cls)}`."
-                )
+        for cls in frozenset(get_origins(*classes)):
+            try:
+                _, current_mode = self.__data[cls]
+
+            except KeyError:
+                pass
+
+            else:
+                if mode == current_mode:
+                    raise RuntimeError(
+                        f"An injectable already exists for the class `{format_type(cls)}`."
+                    )
+
+                elif rank < modes.index(current_mode):
+                    continue
+
+            yield cls
 
 
 """
@@ -315,15 +330,6 @@ Module
 """
 
 
-class ModulePriority(Enum):
-    HIGH = auto()
-    LOW = auto()
-
-    @classmethod
-    def get_default(cls):
-        return cls.LOW
-
-
 @dataclass(repr=False, eq=False, frozen=True, slots=True)
 class Module(EventListener, Broker):
     name: str = field(default=None)
@@ -359,7 +365,7 @@ class Module(EventListener, Broker):
 
     @property
     def __brokers(self) -> Iterator[Broker]:
-        yield from frozen_collection(self.__modules)
+        yield from tuple(self.__modules)
         yield self.__container
 
     def injectable(
@@ -369,14 +375,14 @@ class Module(EventListener, Broker):
         *,
         cls: type[Injectable] = NewInjectable,
         inject: bool = True,
-        on: type | Types = None,
-        override: bool = False,
+        on: type | _Types = None,
+        mode: _Mode = "normal",
     ):
         def decorator(wp):
             factory = self.inject(wp, return_factory=True) if inject else wp
             injectable = cls(factory)
             classes = find_types(wp, on)
-            self.update(classes, injectable, override)
+            self.update(classes, injectable, mode)
             return wp
 
         return decorator(wrapped) if wrapped else decorator
@@ -385,7 +391,11 @@ class Module(EventListener, Broker):
 
     def should_be_injectable(self, wrapped: type = None, /):
         def decorator(wp):
-            self[wp] = ShouldBeInjectable(wp)
+            self.update(
+                (wp,),
+                ShouldBeInjectable(wp),
+                mode="fallback",
+            )
             return wp
 
         return decorator(wrapped) if wrapped else decorator
@@ -393,16 +403,16 @@ class Module(EventListener, Broker):
     def set_constant(
         self,
         instance: _T,
-        on: type | Types = None,
+        on: type | _Types = None,
         *,
-        override: bool = False,
+        mode: _Mode = "normal",
     ) -> _T:
         cls = type(instance)
         self.injectable(
             lambda: instance,
             inject=False,
             on=(cls, on),
-            override=override,
+            mode=mode,
         )
         return instance
 
@@ -418,13 +428,18 @@ class Module(EventListener, Broker):
                 wp.__init__ = self.inject(wp.__init__)
                 return wp
 
-            wrapper = InjectedFunction(wp).update(self)
-            self.add_listener(wrapper)
-            return wrapper
+            function = InjectedFunction(wp)
+
+            @function.on_setup
+            def listen():
+                function.update(self)
+                self.add_listener(function)
+
+            return function
 
         return decorator(wrapped) if wrapped else decorator
 
-    def get_instance(self, cls: type[_T], none: bool = True) -> _T | None:
+    def get_instance(self, cls: type[_T], *, none: bool = True) -> _T | None:
         try:
             injectable = self[cls]
         except KeyError as exc:
@@ -436,23 +451,29 @@ class Module(EventListener, Broker):
         instance = injectable.get_instance()
         return cast(cls, instance)
 
-    def get_lazy_instance(self, cls: type[_T]) -> Lazy[_T | None]:
-        return Lazy(lambda: self.get_instance(cls))
+    def get_lazy_instance(
+        self,
+        cls: type[_T],
+        *,
+        cache: bool = False,
+    ) -> Invertible[_T | None]:
+        if cache:
+            return Lazy(lambda: self.get_instance(cls))
+
+        function = self.inject(lambda instance=None: instance)
+        function.set_owner(cls)
+        return SimpleInvertible(function)
 
     def update(
         self,
         classes: Iterable[type],
         injectable: Injectable,
-        override: bool = False,
+        mode: _Mode = "normal",
     ):
-        self.__container.update(classes, injectable, override)
+        self.__container.update(classes, injectable, mode)
         return self
 
-    def use(
-        self,
-        module: Module,
-        priority: ModulePriority = ModulePriority.get_default(),
-    ):
+    def use(self, module: Module, *, priority: _Priority = "low"):
         if module is self:
             raise ModuleError("Module can't be used by itself.")
 
@@ -482,13 +503,14 @@ class Module(EventListener, Broker):
     def use_temporarily(
         self,
         module: Module,
-        priority: ModulePriority = ModulePriority.get_default(),
+        *,
+        priority: _Priority = "low",
     ) -> ContextManager | ContextDecorator:
-        self.use(module, priority)
+        self.use(module, priority=priority)
         yield
         self.stop_using(module)
 
-    def change_priority(self, module: Module, priority: ModulePriority):
+    def change_priority(self, module: Module, priority: _Priority):
         event = ModulePriorityUpdated(self, module, priority)
 
         with self.notify(event):
@@ -496,11 +518,13 @@ class Module(EventListener, Broker):
 
         return self
 
-    @synchronized
+    @synchronized()
     def unlock(self):
         for broker in self.__brokers:
             broker.unlock()
 
+        return self
+
     def add_listener(self, listener: EventListener):
         self.__channel.add_listener(listener)
         return self
@@ -525,8 +549,8 @@ class Module(EventListener, Broker):
         if self.is_locked:
             raise ModuleLockError(f"`{self}` is locked.")
 
-    def __move_module(self, module: Module, priority: ModulePriority):
-        last = priority == ModulePriority.LOW
+    def __move_module(self, module: Module, priority: _Priority):
+        last = priority != "high"
 
         try:
             self.__modules.move_to_end(module, last=last)
@@ -612,52 +636,59 @@ class Arguments(NamedTuple):
 
 
 class InjectedFunction(EventListener):
-    __slots__ = ("__dict__", "__wrapper", "__dependencies", "__owner")
+    __slots__ = (
+        "__dict__",
+        "__signature__",
+        "__wrapped__",
+        "__dependencies",
+        "__owner",
+        "__setup_queue",
+    )
 
     def __init__(self, wrapped: Callable[..., Any], /):
-        update_wrapper(self, wrapped)
-        self.__signature__ = Lazy[Signature](
-            lambda: inspect.signature(wrapped, eval_str=True)
-        )
+        try:
+            variables = vars(wrapped)
+        except TypeError:
+            pass
+        else:
+            self.__update_vars(variables)
+            del variables
 
-        @wraps(wrapped)
-        def wrapper(*args, **kwargs):
-            args, kwargs = self.bind(args, kwargs)
-            return wrapped(*args, **kwargs)
-
-        self.__wrapper = wrapper
+        update_wrapper(self, wrapped, updated=())
         self.__dependencies = Dependencies.empty()
         self.__owner = None
+        self.__setup_queue = LimitedQueue[Callable[[], Any]]()
+        self.on_setup(self.__set_signature)
 
     def __repr__(self) -> str:
-        return repr(self.__wrapper)
+        return repr(self.wrapped)
 
     def __str__(self) -> str:
-        return str(self.__wrapper)
+        return str(self.wrapped)
 
     def __call__(self, /, *args, **kwargs) -> Any:
-        return self.__wrapper(*args, **kwargs)
+        for function in self.__setup_queue:
+            function()
 
-    def __get__(self, instance: object | None, owner: type):
+        args, kwargs = self.bind(args, kwargs)
+        return self.wrapped(*args, **kwargs)
+
+    def __get__(self, instance: object = None, owner: type = None):
         if instance is None:
             return self
 
-        return self.__wrapper.__get__(instance, owner)
+        return MethodType(self, instance)
 
     def __set_name__(self, owner: type, name: str):
-        if self.__dependencies.are_resolved:
-            raise TypeError(
-                "`__set_name__` is called after dependencies have been resolved."
-            )
-
-        if self.__owner:
-            raise TypeError("Function owner is already defined.")
-
-        self.__owner = owner
+        self.set_owner(owner)
 
     @property
     def signature(self) -> Signature:
-        return self.__signature__()
+        return self.__signature__
+
+    @property
+    def wrapped(self) -> Callable[..., Any]:
+        return self.__wrapped__
 
     def bind(
         self,
@@ -671,21 +702,35 @@ class InjectedFunction(EventListener):
             return Arguments(args, kwargs)
 
         bound = self.signature.bind_partial(*args, **kwargs)
-        dependencies = self.__dependencies.arguments
-        bound.arguments = dependencies | bound.arguments
-
+        bound.arguments = (
+            bound.arguments | self.__dependencies.arguments | bound.arguments
+        )
         return Arguments(bound.args, bound.kwargs)
 
-    def update(self, module: Module):
-        with thread_lock:
-            self.__dependencies = Dependencies.resolve(
-                self.signature,
-                module,
-                self.__owner,
+    def set_owner(self, owner: type):
+        if self.__dependencies.are_resolved:
+            raise TypeError(
+                "Function owner must be assigned before dependencies are resolved."
             )
 
+        if self.__owner:
+            raise TypeError("Function owner is already defined.")
+
+        self.__owner = owner
         return self
 
+    @synchronized()
+    def update(self, module: Module):
+        self.__dependencies = Dependencies.resolve(self.signature, module, self.__owner)
+        return self
+
+    def on_setup(self, wrapped: Callable[[], Any] = None, /):
+        def decorator(wp):
+            self.__setup_queue.add(wp)
+            return wp
+
+        return decorator(wrapped) if wrapped else decorator
+
     @singledispatchmethod
     def on_event(self, event: Event, /):
         pass
@@ -695,3 +740,19 @@ class InjectedFunction(EventListener):
     def _(self, event: ModuleEvent, /) -> ContextManager:
         yield
         self.update(event.on_module)
+
+    def __set_signature(self):
+        self.__signature__ = inspect.signature(self.wrapped, eval_str=True)
+        return self
+
+    def __update_vars(self, variables: Mapping[str, Any], /):
+        def is_dunder(var: str) -> bool:
+            return var.startswith("__") and var.endswith("__")
+
+        restricted_vars = frozenset(var for var in dir(self) if not is_dunder(var))
+        variables = (
+            (var, value)
+            for var, value in variables.items()
+            if var not in restricted_vars
+        )
+        vars(self).update(variables)
